@@ -1,9 +1,9 @@
-;;; quilt.el v0.45.4 - a minor mode for working with files in quilt
+;;; quilt.el v0.46.2 - a minor mode for working with files in quilt
 ;;; http://stakeuchi.sakura.ne.jp/dev/quilt-el
 ;;;
 ;;; Copyright 2005  Matt Mackall <mpm@selenic.com>
 ;;;
-;;; Satoru Takeuchi<nqm08501@nifty.com> took over this package
+;;; Satoru Takeuchi<satoru.takeuchi@gmail.com> took over this package
 ;;; from Matt Mackall.
 ;;;
 ;;; This software may be used and distributed according to the terms
@@ -20,8 +20,20 @@
   (if (> (call-process "quilt" nil nil nil "applied") 0) 1))
 
 (defun quilt-patches-directory ()
-  (or (getenv "QUILT_PATCHES")
-      "patches"))
+  (or (save-excursion
+        (set-buffer (generate-new-buffer " *cmd"))
+        (shell-command
+         (concat "test -f ~/.quiltrc && . ~/.quiltrc ;"
+                 "echo -n $QUILT_PATCHES")
+         t)
+        (unwind-protect
+            (let ((v (buffer-string)))
+              (if (string= "" (buffer-string))
+                  nil
+                v))
+          (kill-buffer (current-buffer))))
+      (or (getenv "QUILT_PATCHES")
+          "patches")))
 
 (defun quilt-find-dir (fn)
   "find the top level dir for quilt from fn"
@@ -36,7 +48,8 @@
   (quilt-find-dir (if fn fn
 		    (let ((fn2 (quilt-buffer-file-name-safe)))
 		      (if fn2 fn2
-			(expand-file-name default-directory))))))
+			(if default-directory
+			    (expand-file-name default-directory)))))))
 
 (defun quilt-drop-dir (fn)
   (let ((d (quilt-find-dir fn)))
@@ -68,8 +81,9 @@
     (if (not qd)
 	(shell-command (concat "quilt " cmd) buf)
       (cd qd)
-      (shell-command (concat "quilt " cmd) buf)
-      (cd d))))
+      (unwind-protect ; make sure to cd back even if an erro occurs.
+	  (shell-command (concat "quilt " cmd) buf)
+	(cd d)))))
 
 (defun quilt-cmd-to-string (cmd)
   "execute a quilt command at the top of the quilt tree for the given buffer"
@@ -78,23 +92,43 @@
     (if (not qd)
 	nil
       (cd qd)
-      (let ((r (shell-command-to-string (concat "quilt " cmd))))
-	(cd d) r))))
+      (unwind-protect ; make sure to cd back even if an error occurs.
+	  (shell-command-to-string (concat "quilt " cmd))
+	(cd d)))))
+
+(defun quilt-cmd-to-list (cmd)
+  "execute a quilt command at the top of the quilt tree for the current buffer.
+Return the lines of the command output as elements of a list."
+  (let ((s (quilt-cmd-to-string cmd)))
+    (if s
+        (split-string s "\n" t))))
 
 (defun quilt-applied-list ()
-  (let ((s (quilt-cmd-to-string "applied")))
-    (if s
-	(split-string s "\n"))))
+  (quilt-cmd-to-list "applied"))
 
 (defun quilt-file-list ()
-  (let ((s (quilt-cmd-to-string "files")))
-    (if s
-	(split-string s "\n"))))
+  (quilt-cmd-to-list "files"))
 
 (defun quilt-patch-list ()
-  (let ((s (quilt-cmd-to-string "series")))
-     (if s
-	 (split-string s "\n"))))
+  (quilt-cmd-to-list "patches"))
+
+(defun quilt-files-affected (&optional first last)
+  (let ((qd (quilt-dir))
+        files fp)
+    (when qd
+      (setq files (quilt-cmd-to-list
+                   (if last
+                       (if (equal first last)
+                           (concat "files " first)
+                         (concat "files --combine " first last))
+                     (if first
+                         (concat "files --combine " first)
+                       "files"))))
+      (setq fp files)
+      (while fp
+        (setcar fp (concat qd (car fp)))
+        (setq fp (cdr fp)))
+      files)))
 
 (defun quilt-top-patch ()
   (if (quilt-bottom-p)
@@ -103,28 +137,31 @@
      (substring (quilt-cmd-to-string "top")  0 -1))))
 
 (defun quilt-complete-list (p l)
-  (defun to-alist (list n)
-    (if list
-	(cons (cons (car list) n)
-	      (to-alist (cdr list) (+ 1 n)))
-      nil))
-  (completing-read p (to-alist l 0) nil t))
+  ; First turn the list into an alist using indices as keys.
+  ; This modifies the list passed as an argument.
+  (let ((list l)
+	(n 0))
+    (while list
+      (setcar list (cons (car list) n))
+      (setq list (cdr list))
+      (setq n (1+ n))))
+  (completing-read p l nil t))
 
 (defun quilt-editable (f)
   (let ((qd (quilt-dir))
-	 (fn (quilt-drop-dir f)))
-    (defun editable (file dirs)
-      (if (car dirs)
-	  (if (file-exists-p (concat qd ".pc/" (car dirs) "/" file))
-	      't
-	    (editable file (cdr dirs)))
-	nil))
+	 (fn (quilt-drop-dir f))
+	 dirs result)
     (if qd
 	(if (quilt-bottom-p)
 	    (quilt-cmd "applied")	; to print error message
-	  (editable fn (if quilt-edit-top-only
-			   (list (quilt-top-patch))
-			 (cdr (cdr (directory-files (concat qd ".pc/"))))))))))
+	  (setq dirs (if quilt-edit-top-only
+			 (list (quilt-top-patch))
+			 (cdr (cdr (directory-files (concat qd ".pc/"))))))
+	  (while (and (not result) dirs)
+	    (if (file-exists-p (concat qd ".pc/" (car dirs) "/" fn))
+		(setq result t)
+	      (setq dirs (cdr dirs))))
+	  result))))
 
 (defun quilt-short-patchname ()
   (let ((p (quilt-top-patch)))
@@ -144,29 +181,34 @@
 	(concat " Q:" (quilt-short-patchname)))
   (force-mode-line-update))
 
-(defun quilt-revert ()
-  (defun revert-or-hook-buffer ()
-    ;; If the file doesn't exist on disk it can't be reverted, but we
-    ;; need the revert hooks to run anyway so that the buffer's
-    ;; editability will update.
-    (if (file-exists-p buffer-file-name)
-	(revert-buffer 't 't)
-      (run-hooks 'after-revert-hook)))
-  (defun revert (buf)
-    (save-excursion
-      (set-buffer buf)
-      (let* ((fn (quilt-buffer-file-name-safe)))
-	(if (quilt-p fn)
-	  (quilt-update-modeline))
-	(if (and (quilt-owned-p fn)
-		 (not (buffer-modified-p)))
-	  (revert-or-hook-buffer)))))
-  (defun revert-list (buffers)
-    (if (not (car buffers))
-	nil
-      (revert (car buffers))
-      (revert-list (cdr buffers))))
-  (revert-list (buffer-list)))
+(defun quilt-revert (filelist)
+  "Refresh contents, editability and modeline of buffers.
+The buffers won't be touched unless their file is a child of the
+current quilt directory. The filelist should contain the absolute
+file names of those files affected by the latest quilt
+operation. Their buffers get reverted to update their
+contents. Other buffers will only get their modeline and
+editability adjusted."
+  (let ((qd (quilt-dir))
+	fn)
+    (dolist (buf (buffer-list))
+      (if (not (string-match "^ " (buffer-name buf)))
+	(save-excursion
+	  (set-buffer buf)
+	  (setq fn (quilt-buffer-file-name-safe))
+	  (when (string-equal qd (quilt-dir))
+	    (quilt-update-modeline)
+	    (when (and (not (buffer-modified-p))
+		       (quilt-owned-p fn))
+	      ;; If the file doesn't exist on disk it can't be reverted, but we
+	      ;; need the revert hooks to run anyway so that the buffer's
+	      ;; editability will update. Files not affected by the latest change
+	      ;; (as listed in filelist) don't need to get reverted either.
+	      (if (and (file-exists-p buffer-file-name)
+		       (member fn filelist))
+		  (progn
+		    (revert-buffer t t t))
+		(run-hooks 'after-revert-hook)))))))))
 
 (defun quilt-push (arg)
   "Push next patch, force with prefix arg"
@@ -175,7 +217,7 @@
   (if arg
       (quilt-cmd "push -f" "*quilt*")
     (quilt-cmd "push -q"))
-  (quilt-revert))
+  (quilt-revert (quilt-files-affected)))
 
 (defun quilt-pop (arg)
   "Pop top patch, force with prefix arg"
@@ -184,25 +226,27 @@
   (if arg
       (quilt-cmd "pop -f")
     (quilt-cmd "pop -q"))
-  (quilt-revert))
+  (quilt-revert (quilt-files-affected)))
 
 (defun quilt-push-all (arg)
   "Push all remaining patches"
   (interactive "P")
   (quilt-save)
-  (if arg
-      (quilt-cmd "push -f" "*quilt*")
-    (quilt-cmd "push -qa"))
-  (quilt-revert))
+  (let ((next (car (quilt-cmd-to-list "next"))))
+    (if arg
+	(quilt-cmd "push -f" "*quilt*")
+      (quilt-cmd "push -qa"))
+    (quilt-revert (quilt-files-affected next))))
 
 (defun quilt-pop-all (arg)
   "Pop all applied patches, force with prefix arg"
   (interactive "P")
   (quilt-save)
-  (if arg
-      (quilt-cmd "pop -af")
-    (quilt-cmd "pop -qa"))
-  (quilt-revert))
+  (let ((fa (quilt-files-affected "-")))
+    (if arg
+	(quilt-cmd "pop -af")
+      (quilt-cmd "pop -qa"))
+    (quilt-revert fa)))
 
 (defun quilt-goto ()
   "Go to a specified patch"
@@ -213,11 +257,20 @@
       (let ((arg (quilt-complete-list "Goto patch: " (quilt-patch-list))))
 	(if (string-equal arg "")
 	    (message "no patch name is supplied")
-	    (quilt-save)
+	  (quilt-save)
+	  (let (cmd first last)
 	    (if (file-exists-p (concat qd ".pc/" arg))
-		(quilt-cmd (concat "pop -q " arg) "*quilt*")
-	      (quilt-cmd (concat "push -q " arg) "*quilt*"))
-	    (quilt-revert))))))
+		(progn
+		  (setq cmd "pop")
+		  (setq first (car (quilt-cmd-to-list (concat "next " arg))))
+		  (setq last (quilt-top-patch)))
+	      (setq cmd "push")
+	      (setq last arg)
+	      (setq first (if (quilt-bottom-p)
+			      "-"
+			    (car (quilt-cmd-to-list "next")))))
+	    (quilt-cmd (concat cmd " -q " arg) "*quilt*")
+	  (quilt-revert (quilt-files-affected first last))))))))
 
 (defun quilt-top ()
   "Display topmost patch"
@@ -265,7 +318,7 @@
       (message "no patch name is supplied")
     (quilt-save)
     (quilt-cmd (concat "new " f ".patch"))
-    (quilt-revert)))
+    (quilt-revert nil)))
 
 (defun quilt-applied ()
   "Show applied patches"
@@ -285,7 +338,7 @@
 	(quilt-cmd (concat "push")))	; to print error message
        (t
 	(quilt-cmd (concat "add " (quilt-drop-dir fn)))
-	(quilt-revert))))))
+	(quilt-revert (list fn)))))))
 
 (defun quilt-edit-patch ()
   "Edit the topmost patch"
@@ -344,7 +397,7 @@
 	  (if (y-or-n-p (format "Really drop %s? " dropped))
 	      (progn
 		(quilt-cmd (concat "remove " dropped))
-		(quilt-revert)))))))))
+		(quilt-revert (list f))))))))))
 
 (defun quilt-edit-series ()
   "Edit the patch series file"
@@ -385,10 +438,10 @@
 	  (if (not p)
 	      (quilt-cmd "applied")	; to print error message
 	    (if (y-or-n-p (format "Really delete %s?" p))
-		(progn
+		(let ((fa (quilt-files-affected p p)))
 		  (quilt-save)
 		  (quilt-cmd (concat "delete " p))
-		  (quilt-revert)))))))))
+		  (quilt-revert fa)))))))))
 
 (defun quilt-header-commit ()
   "commit to change patch header"
